//=============================================================================
//  MuseScore
//  Linux Music Score Editor
//  $Id:$
//
//  Explode Parts plugin
//
//  Copyright (C)2013 Colin Whyles
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License version 2.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//=============================================================================

//---------------------------------------------------------
// Known bugs:
// Ties are not handled properly. Even though they are copied in the code, they do not appear
// and do not get carried through to the MuseScore file.
//
// Instrument families ought to be read from files for ease of user modification.
//
// The synthesizer does not seem to pickup the correct instruments until it is refreshed 
// by re-selecting one of the instruments.
//
// The makeNewScore flag does not work. The new score is created but the notes are not
// copying across. Leave it at false. Mark it 'under development'!
//
// Does not cater for multiple voices in a staff.
//
// Does not copy leading notes if not starting on a bar line.
//
// Hint:
// When arranging to transposing instruments, begin with 'Concert Pitch' toggled ON.
// The keys the transposing instruments are in will not be shown until after you have
// exploded the parts and toggled Concert Pitch OFF. This might seem counter-intuitive,
// but it works. If you begin the arrangement with Concert Pitch OFF, then transposing 
// instruments will be off-key.
//
// Possible development:
// Introduce a degree of randomness in choosing instrument families so that arrangements 
// are not always predictable.
//---------------------------------------------------------

// Globals
var pluginName = qsTr("Explode Parts");
var IFSelnForm;					// The instrument family selection form
var makeNewScore = false;		// Whether to make a new score or add to the current one
var noMeasures;					// Number of measures required
var chordList = [];				// Create array to store each chord. 
								// Thanks to underquark for the crafty way of using the array.
var staves;						// initial number of staves
var readStaff = 0;				// The staff we are reading from
var tickList = [];				// Where the chords and rests fall. 
								// We need this because the cursor in MuseScore doesn't seem 
								// to only act on a single voice. It acts on the smallest division among all voices.
var noVoices = 0;				// Number of voices in the staff

//---------------------------------------------------------
//    init
//---------------------------------------------------------

function init() {
};

//---------------------------------------------------------
//    close
//---------------------------------------------------------

function close() {
};

//---------------------------------------------------------
// uniqid() from Kevin van Zonneveld/Kankrelune
// We use this to provide unique names for instruments that are needed beyond our named lists.
// The names are only used internally. 
//---------------------------------------------------------

function uniqid (prefix, more_entropy) {
  // +   original by: Kevin van Zonneveld (http://kevin.vanzonneveld.net)
  // +    revised by: Kankrelune (http://www.webfaktory.info/)
  // %        note 1: Uses an internal counter (in php_js global) to avoid collision
  // *     example 1: uniqid();
  // *     returns 1: 'a30285b160c14'
  // *     example 2: uniqid('foo');
  // *     returns 2: 'fooa30285b1cd361'
  // *     example 3: uniqid('bar', true);
  // *     returns 3: 'bara20285b23dfd1.31879087'
  if (typeof prefix === 'undefined') {
    prefix = "";
  }

  var retId;
  var formatSeed = function (seed, reqWidth) {
    seed = parseInt(seed, 10).toString(16); // to hex str
    if (reqWidth < seed.length) { // so long we split
      return seed.slice(seed.length - reqWidth);
    }
    if (reqWidth > seed.length) { // so short we pad
      return Array(1 + (reqWidth - seed.length)).join('0') + seed;
    }
    return seed;
  };

  // BEGIN REDUNDANT
  if (!this.php_js) {
    this.php_js = {};
  }
  // END REDUNDANT
  if (!this.php_js.uniqidSeed) { // init seed with big random int
    this.php_js.uniqidSeed = Math.floor(Math.random() * 0x75bcd15);
  }
  this.php_js.uniqidSeed++;

  retId = prefix; // start with prefix, add current milliseconds hex string
  retId += formatSeed(parseInt(new Date().getTime() / 1000, 10), 8);
  retId += formatSeed(this.php_js.uniqidSeed, 5); // add seed hex string
  if (more_entropy) {
    // for more entropy we add a float lower to 10
    retId += (Math.random() * 10).toFixed(8).toString();
  }

  return retId;
}

//---------------------------------------------------------
// stavesRequired
//	Find out how many parts/staves are required
// 	by scanning chords and finding maximum number of notes
//---------------------------------------------------------

function stavesRequired(chords) {
	var x = 0;
	
	for (i = 0; i < chords.length; i++) {
		if (x < chords[i].notes) 
			x = chords[i].notes;
	}
	return x;
}

//---------------------------------------------------------
// scanForVoices
// This scans the selection looking to see how many voices there are.
//---------------------------------------------------------
function scanForVoices (score) {
	var voiceCnt = 0;
	var cursor = new Cursor(score);
	var endCsr = new Cursor(score);
	cursor.rewind();
	endCsr.rewind();
	endCsr.goToSelectionEnd();
	// There can only be four voices on a staff.
	for (var voice = 0; voice < 4; voice++) {
		cursor.goToSelectionStart();
		cursor.voice = voice;
		// Look for a chord. There only has to be one to require a voice.
		while (cursor.tick() < endCsr.tick())
		{
			if (cursor.isChord())
			{
				voiceCnt++;
				break;
			}
			cursor.next();
		}
	}
	return voiceCnt;
}

//---------------------------------------------------------
// initParts()
// Scan existing staves to find out what parts exist.
// Return an array of the part index, number of staves and instrument name.
//---------------------------------------------------------
function initParts() {
	var partList = [];
	var ptStaves = 0;
	var lname = "";
	for (var x = 0; x < curScore.parts; x++) {
		ptStaves = curScore.part(x).staves;		// Number of staves for this part
		lname = curScore.part(x).longName;
		// catch the Unicode flat symbol
		// MuseScore will sort out the Unicode conversion, but prefers us to pass it in as 'b'
		// This should perhaps be more generic and scan the whole string, 
		// but I think this should catch most if not all cases.
		if (lname.charAt(2) == ' ' && lname.charCodeAt(1) == 0x266D)
			lname = lname .substr(0,1)+'b'+lname.substr(2,lname.length-2);
	
		partList[x] = [x, ptStaves, lname];
	}
	return partList;
}

//---------------------------------------------------------
// getInstrName()
// This first searches for a name from our named list. If one is not found then
// a unique name is generated to keep the parts separate.
//---------------------------------------------------------

function getInstrName(voice, family, family_ch) {
	var instr = family[family_ch][voice];
	if (instr == "") {
		// Then create a unique name
		instr = uniqid("instr");
	}
	return instr;
}

// When the user has selected the instrument family, the form control continues here
// The first voice has already been read into the array.
function newParts() {
	var title = curScore.title;
	
	// Instrument families
	// The parts are copied into sets of instruments from each family.
	// The user has chosen the family to use.
	// Currently the instruments are used top-down from each array. 
	// It might be an interesting development to select variations of instrument families 
	// at random in order to provide variety to arrangements.
	// Provision has been made for eight instruments in each family simply to prevent overflow
	// if we are exploding big chords. Let us hope no-one uses bigger chords than this!
	// A basic, rather arbitrary, overflow check has been made in code.
	// These arrays must always end with an empty string in order for successive
	// voices to have unique name generated.
	var noFamily = [ "", "", "", "", "", "", "", "", "", "" ];
	var stringFamily = [ "Violins 1", "Violins 2", "Violas", "Violoncellos", "Double Basses","","","" ];
	var woodwindFamily = [ "Flute", "Oboe", "Bb Clarinet", "Bassoon", "English Horn","","","" ];
	var brassFamily = [ "Bb Trumpet", "Horn", "Trombone", "Tuba", "Bb Cornet", "Flugelhorn","","" ];
	var stringQuartet = [ "Violin", "Violins 2", "Viola", "Violoncello","","","","","" ];
	var family = [ noFamily, stringFamily, woodwindFamily, brassFamily, stringQuartet ];
	var family_ch = 0;		// default to no instrument family chosen
	var partList = [];		// We keep a list of parts/staves here by recording their index in [0], 
							// number of staves in [1] and instrument name in [2].
	var voicePartNdx = [];	// Indices of the parts of the current voice.
	var voiceSeparation = true;
	var voiceIn = 0;
	var stavesReq = 0;
	var voiceCnt = 0;		// Keeps track of number of voices in all parts
	
	//---------------------------------------------------------
	// copyVoice()
	// Copy voice from chordList array to new part
	//---------------------------------------------------------
	function copyVoice(cLndx, voiceIn, stavesReq, writeToCsr) {
		var chrd;
		var instr = "";
		var tick = 0;
	
		chrd = chordList[cLndx];
		var chdVoice = 0;
		if (typeof chrd === 'object') {
			function copyNoteToPart(pCursor, chrd, chVoice, staff) {
				var nt = new Note();
				var chd = new Chord();
				var cursr = new Cursor(curScore);
				cursr.rewind();
				cursr = pCursor;
				nt.pitch = chrd.note(chVoice).pitch;					// pick out the note for this voice
				nt.tied = chrd.note(chVoice).tied;						// it might be tied
				nt.tpc = chrd.note(chVoice).tpc;						// tonal pitch class - e.g. enharmonic
				nt.userAccidental = chrd.note(chVoice).userAccidental;	// user added accidental
				chd.tickLen = chrd.tickLen;								// add the length
				chd.type = chrd.type;									// add the type
				chd.addNote(nt);										// add it to our one-note chrd
				cursr.staff = staff;									// go to the stave
				cursr.add(chd);											// add it
			}
			// copy notes to parts one by one
			for ( ; chdVoice < chrd.notes; chdVoice++ ) {
				//	Here is where we need to select the staff according to whether they pre-existed or not.
				copyNoteToPart(writeToCsr, chrd, chdVoice, voicePartNdx[stavesReq - 1 - chdVoice] );
			}
			// if empty voices recycle the notes
			while (chdVoice < stavesReq) {
				for (reUseVce = 0; chdVoice < stavesReq && reUseVce < chrd.notes; reUseVce++) {
					//	Ditto.
					copyNoteToPart(writeToCsr, chrd, reUseVce, voicePartNdx[stavesReq - 1 - chdVoice] );
					chdVoice++;
				}
			}			
		} else {
			// add rest if it's not a chord
			while (chdVoice < stavesReq) {
				// Using a function avoids incrementing writeToCsr
				function addRestToPart(pCursor, chrd, staff) {
					var cursr = new Cursor(curScore);
					cursr.rewind();
					cursr = pCursor;
					rest = new Rest();
					rest.tickLen = chrd;
					cursr.staff = staff;
					cursr.voice = 0;
					cursr.add(rest);
					pCursor.staff = readStaff;							// reset read cursor.
				}
				instr = family[family_ch][stavesReq - 1 - chdVoice];			// get the instrument name from the family name list.
				addRestToPart(writeToCsr, chrd, voicePartNdx[chdVoice] /*getStaffbyName(instr)*/);
				chdVoice++;
			}
		}
	}	// end function copyVoice()

	partList = initParts();			// make a list of existing staves.
	
	//------------------------------------------------
	// getStaffbyName()
	// scan partList looking for an instrument by name
	// Success: return staff index
	// Failure: return false
	//------------------------------------------------
	function getStaffbyName(instrName) {
		var staveCnt = 0;
		for (var x = 0; x < partList.length; x++) {
			if (partList[x][2] == instrName)
				return staveCnt;
			staveCnt += partList[x][1];		// Keep a running total of staves
		}
		return false;		// Should be safe as we should never be writing to staff 0.
	}	// end getStaffbyName()
	
	curScore.startUndo();
	
	if (IFSelnForm) {
		family_ch = getFamily(IFSelnForm);
		voiceSeparation = getVoiceSeparation();
	}
	if (makeNewScore) {
		var partScore = new Score();
		partScore.startUndo();
		partScore.title = title.concat(" Part Explosion");
		staves = 0;		// Do not need to offset for existing staves
	} else {
		partScore = curScore;
	}
	
	noVoices = 1;
	if (voiceSeparation)
		noVoices = scanForVoices(curScore);		// This needs fixing to pass partScore or curScore as appropriate.
	
	voiceIn = 0;
	voiceCnt = 0;
	while (voiceIn < noVoices) {
		readVoice(voiceIn);
		// Find out how many staves we need for the chord voices.
		stavesReq = stavesRequired(chordList);
		if (stavesReq == 0) {
			QMessageBox.warning(this, "No staves required", qsTr("<b>No staves required.</b>"));
			curScore.endUndo();
			return;		// No selection made.
		}
		//	Here, we determine whether we already have the staves requested.
		//	An array of indices and names is used, partList[].
		var instr = "";
		// By referencing stavesReq this implies we are working in one voice
		// to the exclusion of all others.
		voicePartNdx = [];		// Clear the voice-part cross reference array
		for ( var cnt = 0; cnt < stavesReq && cnt < 8; cnt++) {	// n < 8 prevents overflow of the family arrays
			instr = getInstrName(voiceCnt, family, family_ch);			// get the instrument name from the family name list.
			
			// Check whether we need a staff
			if (!(pt = getStaffbyName(instr))) {
				// Doesn't exist so create a new staff.
				// Score.appendPart() needs an instrument name or it creates a default staff. Once created, we should then stop using the instrument name and use our staff index from partList.
				partScore.appendPart(instr);		// No, so create a staff for it,
				function getNextPartNdx(partList) {
					var y = 0
					for (x = 0; x < partList.length; x++) {
						y += partList[x][1];
					}
					return y;
				}
				// and add it to the partList array.
				// Note the assumption that we only need 1 staff per part.
				// We do not cater for multi-staff instruments.
				partList[partList.length] = [getNextPartNdx(partList), 1, instr];
			} else {
				// Copy the staff index from pt
				partList[partList.length] = [pt, 1, instr];
			}
				voicePartNdx[voicePartNdx.length] = partList[partList.length-1][0];	// Store a cross reference
				voiceCnt++;		// Keep running total of voices
		}
	
		if (makeNewScore) {
			partScore.keysig = curScore.keysig;
			partScore.appendMeasures(noMeasures);
			var readFromCsr = new Cursor(partScore);
			readFromCsr.rewind();
		} else {
			var readFromCsr = new Cursor(curScore);
			readFromCsr.rewind();
		}
		
		// Copy the notes to the parts
		readFromCsr.goToSelectionStart();
		readFromCsr.voice = voiceIn;		// Copy them from voiceIn
		readFromCsr.staff = readStaff;		// Follow the staff we have read from
	
		var writeToCsr = new Cursor(curScore);
		writeToCsr.rewind();
		writeToCsr.goToSelectionStart();
		writeToCsr.voice = 0;				// Always write to first voice
		tick = 0;
		// Some voices might not start at the selection start
		while (readFromCsr.tick() < tickList[tick])
			readFromCsr.next();	// cursor.add() function increments a cursor, but that was lost inside the function.
		while (writeToCsr.tick() < tickList[tick]) 
			writeToCsr.next();
		while (tick < chordList.length) {
			copyVoice(tick, voiceIn, stavesReq, writeToCsr);
			tick++;
			while (readFromCsr.tick() < tickList[tick])
				readFromCsr.next();	// cursor.add() function increments a cursor, but that was lost inside the function.
			while (writeToCsr.tick() < tickList[tick]) 
				writeToCsr.next();
		}
		
		voiceIn++;
	}	// end while voiceIn
	
	if (makeNewScore)
		partScore.endUndo();
	readFromCsr.rewind();
	chordList = [];		// free the array
	curScore.endUndo();
};

//---------------------------------------------------------
// initFamilySelnForm()
// Initialise and open the form for selecting the family of instruments required.
// The program flow stops once the form is displayed, waiting for the user to select
// a set of instruments and press the OK button.
// Before we showed the form, we connected the OK button to the newParts() function,
// which is where the program continues.
//---------------------------------------------------------

function initFamilySelnForm() {
	var loader = new QUiLoader(null);
	var file   = new QFile(pluginPath + "/explodeparts-Instruments.ui");
	file.open(QIODevice.OpenMode(QIODevice.ReadOnly, QIODevice.Text));
	IFSelnForm = loader.load(file, null);
	if (IFSelnForm) {
		IFSelnForm.buttonBox.accepted.connect(newParts);	// This tells the Qt form to continue with function newParts if OK is clicked
															// It uses the Qt concept of connecting the 'accepted' button to the function newParts()
    	IFSelnForm.buttonBox.rejected.connect(close);		// Similarly, this tells the form to call the close() function if cancelled
		IFSelnForm.show();
    }
}

//---------------------------------------------------------
// getFamily()
// Find which family of instruments has been chosen
// We probably do not use Qt the way it was intended. 
// Rather than use its signal and slot mechanism,
// we simply scan the widgets.
//---------------------------------------------------------

function getFamily() {
	if (IFSelnForm.noSelection.checked == true) 
  		return 0;
	if (IFSelnForm.stringSeln.checked == true) 
  		return 1;
	if (IFSelnForm.woodwindSeln.checked == true) 
  		return 2;
	if (IFSelnForm.brassSeln.checked == true) 
  		return 3;
	if (IFSelnForm.quartetSeln.checked == true) 
  		return 4;
  		
  	return 0;	// Just in case!
}

//---------------------------------------------------------
// getVoiceSeparation()
// Find out if the user wants the separate voices processing
//---------------------------------------------------------

function getVoiceSeparation() {
	return IFSelnForm.voiceSeparation.checked;
}

//---------------------------------------------------------
// readVoice()
// Read a voice into the chordList array
//---------------------------------------------------------
function readVoice(voice) {
	var tick = 0;
	staves = curScore.staves;	// initial number of staves
	var cursor = new Cursor(curScore);
	cursor.staff = 0;			// gets set to selected staff when the selection is found.
	cursor.rewind();
	
	//  Bound the selection
	var selnEnd = new Cursor(curScore);
	cursor.goToSelectionStart();
	cursor.voice = voice;
	readStaff = cursor.staff;		// Note the staff we are reading from!
	selnEnd.goToSelectionEnd();
	selnEnd.voice = voice;

	//  Read notes and rests into chordList array
	var tickCnt = cursor.tick();
	tick = 0;
	tickList = [];		// We record the corresponding tick times in tickList.
	chordList = [];
	while (cursor.tick() < selnEnd.tick())
	{
		//  Read in the chords and rests etc.
		if (cursor.isChord()) {
			var chrd = cursor.chord();
			chordList.push(chrd);
			tickList[tick] = cursor.tick();
			tick++;
		} else if (cursor.isRest()) {
			var rest = cursor.rest();                    
			chordList.push(rest.tickLen);
			tickList[tick] = cursor.tick();
			tick++;
		}
		cursor.next();
	}
	cursor.rewind();
}

//---------------------------------------------------------
//  explodeParts()
//  from all chords copy notes to other staves
// 	Strategy: 
// 	1: Scan selected range or whole if no selection for maximum notes per chord.
// 	2: Create that many staves
// 	3: If a chord has less notes, reuse notes for the lower voices.
//---------------------------------------------------------

function explodeParts() {
	//  Bounce if not a proper score
	if (typeof curScore === 'undefined') 
		return; 

	//  Bound the selection
	var cursor = new Cursor(curScore);
	cursor.rewind();
	cursor.goToSelectionEnd();
	if (cursor.eos() == true) {
		QMessageBox.information(this, "No Selection", qsTr("<b>No selection made.</b> <br>Select a range of measures. Explode Parts will create a number of new staves and copy the voices of the chords into them."));
		return;		// No selection made.
	}

	if (scanForVoices(curScore) == 0) {
		// Nothing to do.
		QMessageBox.information(this, "No Voices", qsTr("<b>No voices in selection.</b> <br>Select a range of measures containing notes."));
		return;
	}
	
	// Read the first voice
	//readVoice(0);
	// Note how many measures we need 
	// ?? Do we need this? Only used for partScore but ought to only create the range of measures selected.
	noMeasures = curScore.measures;

	// Display the dialogue form to find out where these notes are to go.
	initFamilySelnForm();
}

var mscorePlugin = {
      menu: 'Plugins.Explode Parts',
      init: init,
      run:  explodeParts,
      onClose: close
      };

mscorePlugin;